#include <l4/re/env.h>
#include <l4/sys/vcon.h>
#include <string.h>

#include <stdlib.h>       /* exit */

/* See: pkg/l4re-core/uclibc/lib/contrib/uclibc/libc/misc/internals/__uClibc_main.c */

static char **envp = 0;
static void (*__app_fini)(void) = 0;
static void (*__rtld_fini)(void) = 0;

void __uClibc_main(int (*main)(int, char **, char **),
                   int argc,
                   char **argv,
                   void (*app_init)(void),
                   void (*app_fini)(void),
                   void (*rtld_fini)(void),
                   void *stack_end)
{
    unsigned long *entry;

    /* The environment is found after argv. */

    envp = &argv[argc + 1];

    /* NOTE: Here, uClibc tests for a lack of environment but later on still
       NOTE: searches an auxiliary vector which may not be present. */

    /* Skip the environment vector which is terminated by a null entry. */

    entry = (unsigned long *) envp;

    while (*entry)
        entry++;

    entry++;

    /* Search the auxiliary vector for the L4 environment entry.
       For some reason, incrementing an ElfW(Auxv) pointer seems to increment
       by only 8 bytes instead of the 16 bytes of the structure size (for 64-bit
       architectures). In fact, the Elf structures are just unusable. So, a
       plain word pointer is used instead. */

    while (*entry)
    {
        /* Search for AT_L4_ENV as the type. */

        if (*entry == 0xf1)
            l4re_global_env = (l4re_env_t *) *(entry + 1);
        entry += 2;
    }

    /* Record finalisation routines. */

    if (app_fini)
      __app_fini = app_fini;

    if (rtld_fini)
      __rtld_fini = rtld_fini;

    /* NOTE: Need to perform __preinit_array processing. */

    /* Run the application's constructors now. */

    if (app_init)
        app_init();

    /* NOTE: Need to perform __init_array processing. */

    exit(main(argc, argv, envp));
}

void __uClibc_fini(void)
{
    /* NOTE: Need to perform __fini_array processing. */

    if (__app_fini)
        (__app_fini)();

    if (__rtld_fini)
        (__rtld_fini)();
}
